'use strict';

Object.defineProperty(exports, "__esModule", {
  value: true
});

var _extends = Object.assign || function (target) { for (var i = 1; i < arguments.length; i++) { var source = arguments[i]; for (var key in source) { if (Object.prototype.hasOwnProperty.call(source, key)) { target[key] = source[key]; } } } return target; };

var _createClass = function () { function defineProperties(target, props) { for (var i = 0; i < props.length; i++) { var descriptor = props[i]; descriptor.enumerable = descriptor.enumerable || false; descriptor.configurable = true; if ("value" in descriptor) descriptor.writable = true; Object.defineProperty(target, descriptor.key, descriptor); } } return function (Constructor, protoProps, staticProps) { if (protoProps) defineProperties(Constructor.prototype, protoProps); if (staticProps) defineProperties(Constructor, staticProps); return Constructor; }; }();

var _Logger = require('./Logger');

var _Logger2 = _interopRequireDefault(_Logger);

var _lodash = require('lodash');

var _lodash2 = _interopRequireDefault(_lodash);

var _yup = require('yup');

var yup = _interopRequireWildcard(_yup);

var _stream = require('stream');

var _stream2 = _interopRequireDefault(_stream);

var _v = require('uuid/v4');

var _v2 = _interopRequireDefault(_v);

function _interopRequireWildcard(obj) { if (obj && obj.__esModule) { return obj; } else { var newObj = {}; if (obj != null) { for (var key in obj) { if (Object.prototype.hasOwnProperty.call(obj, key)) newObj[key] = obj[key]; } } newObj.default = obj; return newObj; } }

function _interopRequireDefault(obj) { return obj && obj.__esModule ? obj : { default: obj }; }

function _asyncToGenerator(fn) { return function () { var gen = fn.apply(this, arguments); return new Promise(function (resolve, reject) { function step(key, arg) { try { var info = gen[key](arg); var value = info.value; } catch (error) { reject(error); return; } if (info.done) { resolve(value); } else { return Promise.resolve(value).then(function (value) { step("next", value); }, function (err) { step("throw", err); }); } } return step("next"); }); }; }

function _classCallCheck(instance, Constructor) { if (!(instance instanceof Constructor)) { throw new TypeError("Cannot call a class as a function"); } }

var log = new _Logger2.default({ component: "EthSyncCursor" });

var MIN_BLOCK_RANGE = 100;
var DEFAULT_DELIMETER = "\n";

var schema = yup.object().shape({
  fromBlock: yup.number().required("Sync cursor missing fromBlock"),
  toBlock: yup.number().min(1).required("Sync cursor missing toBlock"),
  contract: yup.object().required("Sync cursor missing contract to query against"),
  web3: yup.object().required("Sync cursor is missing web3 instance"),
  eventName: yup.string(), //not required
  options: yup.object() //not required
});

var TOO_MANY = /more than \d+ results/;

/**
 * Cursor holds current offset of blocks to retrieve in an iterative way. This
 * allows client code to have control over when to grab next batch of data.
 */

var EthSyncCursor = function () {
  function EthSyncCursor(props) {
    var _this = this;

    _classCallCheck(this, EthSyncCursor);

    schema.validateSync(props);

    //starting block
    this.fromBlock = props.fromBlock;

    //interim end block. This may be different than final end block
    //if block range results in > 1k events
    this.toBlock = props.toBlock;

    //the actual range end block
    this.finalEnd = props.toBlock;

    //web3 instance in case we need to query additional metadata
    this.web3 = props.web3;

    //number of blocks to add when computing next block iteration
    this.increment = 0;

    //the contract we're querying against
    this.contract = props.contract;

    //specific event name we're looking for
    this.eventName = props.eventName;

    //custom delimiter if data contains newlines
    this.recordDelimeter = props.recordDelimeter || DEFAULT_DELIMETER;

    //whether to retrieve transaction receipts for each txn hash
    this.includeReceipt = props.includeReceipt;

    //any filter options to include in the query
    this.options = props.options;

    //bootstrap the process by providing a cached snapshot of
    //event data
    this.snapshotProvider = props.snapshotProvider;

    //metadata about the retrieval process
    this.meta = {
      rpcCalls: 0,
      fromBlock: 0,
      toBlock: 0
    };

    ['init', 'nextBatch', '_pull', '_bootstrap', '_processSnapshotItems', '_getReceipts'].forEach(function (fn) {
      return _this[fn] = _this[fn].bind(_this);
    });
  }

  /**
   * Initialize the cursor by performing the first query of events.
   * @returns Promise that when complete indicates the query has
   * finished and all events were given through the callback
   */


  _createClass(EthSyncCursor, [{
    key: 'init',
    value: function init(callback, badCallback) {
      var _this2 = this;

      return new Promise(function (done, err) {
        //first, try to bootstrap with a snapshot provider
        if (_this2.snapshotProvider) {
          _this2._bootstrap(callback).then(function () {
            _this2._pull(done, err, callback, badCallback);
          }).catch(function (e) {
            return err(e);
          });
        } else {
          //just pull in according to block range
          _this2._pull(done, err, callback, badCallback);
        }
      });
    }

    /*
     * Get the next batch of items.
     * @returns Promise that when complete indicates query has finished.
     */

  }, {
    key: 'nextBatch',
    value: function nextBatch(callback, badHandler) {
      var _this3 = this;

      return new Promise(function (done, err) {
        _this3._pull(done, err, callback, badHandler);
      });
    }
  }, {
    key: '_bootstrap',
    value: function _bootstrap(callback, badHandler) {
      var _this4 = this;

      return new Promise(function () {
        var _ref = _asyncToGenerator( /*#__PURE__*/regeneratorRuntime.mark(function _callee4(done, err) {
          var s, data, start, endHandler, handleData, writeable;
          return regeneratorRuntime.wrap(function _callee4$(_context4) {
            while (1) {
              switch (_context4.prev = _context4.next) {
                case 0:
                  s = null;
                  _context4.prev = 1;
                  _context4.next = 4;
                  return _this4.snapshotProvider.getLatest();

                case 4:
                  s = _context4.sent;
                  _context4.next = 10;
                  break;

                case 7:
                  _context4.prev = 7;
                  _context4.t0 = _context4['catch'](1);
                  return _context4.abrupt('return', err(_context4.t0));

                case 10:
                  if (s) {
                    _context4.next = 12;
                    break;
                  }

                  return _context4.abrupt('return', done());

                case 12:
                  data = "";
                  start = Date.now();

                  endHandler = function () {
                    var _ref2 = _asyncToGenerator( /*#__PURE__*/regeneratorRuntime.mark(function _callee() {
                      var d;
                      return regeneratorRuntime.wrap(function _callee$(_context) {
                        while (1) {
                          switch (_context.prev = _context.next) {
                            case 0:
                              if (!(data.length > 0)) {
                                _context.next = 11;
                                break;
                              }

                              _context.prev = 1;

                              //process last bit forcing send since it's last bit of data
                              d = data;

                              data = "";
                              _context.next = 6;
                              return _this4._processSnapshotItems(callback, d, true);

                            case 6:
                              _context.next = 11;
                              break;

                            case 8:
                              _context.prev = 8;
                              _context.t0 = _context['catch'](1);

                              callback(_context.t0);

                            case 11:
                              log.info("EthSync bootstrapped", _this4.snapCount, "txns in", Date.now() - start, "ms with snapshot containing block range", _this4.snapStart, "-", _this4.fromBlock - 1);
                              done();

                            case 13:
                            case 'end':
                              return _context.stop();
                          }
                        }
                      }, _callee, _this4, [[1, 8]]);
                    }));

                    return function endHandler() {
                      return _ref2.apply(this, arguments);
                    };
                  }();

                  handleData = function () {
                    var _ref3 = _asyncToGenerator( /*#__PURE__*/regeneratorRuntime.mark(function _callee2(buff, encoding, cb) {
                      var idx;
                      return regeneratorRuntime.wrap(function _callee2$(_context2) {
                        while (1) {
                          switch (_context2.prev = _context2.next) {
                            case 0:
                              if (!buff) {
                                _context2.next = 16;
                                break;
                              }

                              data += buff.toString();
                              _context2.prev = 2;
                              _context2.next = 5;
                              return _this4._processSnapshotItems(callback, data);

                            case 5:
                              data = _context2.sent;
                              _context2.next = 13;
                              break;

                            case 8:
                              _context2.prev = 8;
                              _context2.t0 = _context2['catch'](2);
                              idx = data.indexOf(_this4.recordDelimeter);

                              if (idx > 0) {
                                data = data.substring(idx + 1); //skip over bad data
                              }
                              callback(_context2.t0);

                            case 13:
                              //log.debug("Requesting next buffer of data...");
                              cb();
                              _context2.next = 18;
                              break;

                            case 16:
                              _context2.next = 18;
                              return endHandler();

                            case 18:
                            case 'end':
                              return _context2.stop();
                          }
                        }
                      }, _callee2, _this4, [[2, 8]]);
                    }));

                    return function handleData(_x3, _x4, _x5) {
                      return _ref3.apply(this, arguments);
                    };
                  }();

                  writeable = new _stream2.default.Writable({
                    decodeStrings: false,
                    write: handleData
                  });


                  s.pipe(writeable);
                  writeable.on("finish", _asyncToGenerator( /*#__PURE__*/regeneratorRuntime.mark(function _callee3() {
                    return regeneratorRuntime.wrap(function _callee3$(_context3) {
                      while (1) {
                        switch (_context3.prev = _context3.next) {
                          case 0:
                            _context3.next = 2;
                            return endHandler();

                          case 2:
                          case 'end':
                            return _context3.stop();
                        }
                      }
                    }, _callee3, _this4);
                  })));

                case 19:
                case 'end':
                  return _context4.stop();
              }
            }
          }, _callee4, _this4, [[1, 7]]);
        }));

        return function (_x, _x2) {
          return _ref.apply(this, arguments);
        };
      }());
    }
  }, {
    key: '_processSnapshotItems',
    value: function () {
      var _ref5 = _asyncToGenerator( /*#__PURE__*/regeneratorRuntime.mark(function _callee5(callback, data, last) {
        var idx, obj, block, _block;

        return regeneratorRuntime.wrap(function _callee5$(_context5) {
          while (1) {
            switch (_context5.prev = _context5.next) {
              case 0:
                idx = data.indexOf(this.recordDelimeter);

              case 1:
                if (!(idx > 0)) {
                  _context5.next = 17;
                  break;
                }

                obj = data.substring(0, idx);
                block = JSON.parse(obj);

                block.number -= 0;
                if (this.meta.fromBlock === 0) {
                  this.meta.fromBlock = block.number;
                }
                this.meta.toBlock = block.number;
                //log.debug("Calling callback with new block...");
                _context5.next = 9;
                return callback(null, block.transactions, _extends({}, this.meta));

              case 9:
                //log.debug("Callback complete");
                if (block.number && block.number > this.fromBlock) {
                  this.fromBlock = block.number + 1;
                }
                if (block.number && (!this.snapStart || this.snapStart > block.number)) {
                  this.snapStart = block.number;
                }
                if (!this.snapCount) {
                  this.snapCount = 0;
                }
                this.snapCount += block.transactions.length;

                data = data.substring(idx + 1);
                idx = data.indexOf(this.recordDelimeter);
                _context5.next = 1;
                break;

              case 17:
                if (!(data.length > 0 && last)) {
                  _context5.next = 25;
                  break;
                }

                _block = JSON.parse(data);

                _block.number -= 0;
                this.meta.toBlock = _block.number;
                _context5.next = 23;
                return callback(null, _block.transactions, _extends({}, this.meta));

              case 23:
                if (_block.number && _block.number > this.fromBlock) {
                  this.fromBlock = _block.number + 1;
                }
                data = "";

              case 25:
                return _context5.abrupt('return', data);

              case 26:
              case 'end':
                return _context5.stop();
            }
          }
        }, _callee5, this);
      }));

      function _processSnapshotItems(_x6, _x7, _x8) {
        return _ref5.apply(this, arguments);
      }

      return _processSnapshotItems;
    }()
  }, {
    key: '_pull',
    value: function () {
      var _ref6 = _asyncToGenerator( /*#__PURE__*/regeneratorRuntime.mark(function _callee6(done, err, cb, badCb) {
        var span, config, contract, evtName, start, events, byBlock, blocks, meta, i, b, _start, end, newSpan, totalSpan, oldEnd;

        return regeneratorRuntime.wrap(function _callee6$(_context6) {
          while (1) {
            switch (_context6.prev = _context6.next) {
              case 0:
                span = this.toBlock - this.fromBlock;

                if (!(span < 0)) {
                  _context6.next = 4;
                  break;
                }

                log.error("Invalid block range. Start is before end", this.fromBlock, this.toBlock);
                //return err(new Error("Start block is after end block"));
                return _context6.abrupt('return', done(undefined));

              case 4:

                log.debug("Querying", span, "blocks for logs in range", this.fromBlock, "-", this.toBlock);
                config = _extends({}, this.options, {
                  fromBlock: this.fromBlock,
                  toBlock: this.toBlock
                });
                _context6.prev = 6;
                contract = this.contract;
                evtName = this.eventName || "allEvents";
                start = Date.now();

                this.meta.rpcCalls++;
                this.meta.fromBlock = this.fromBlock;
                this.meta.toBlock = this.toBlock;

                //attempt to get events
                _context6.next = 15;
                return contract.getPastEvents(evtName, config);

              case 15:
                events = _context6.sent;


                //always make sure events are sorted by block and txn index
                events.sort(function (a, b) {
                  var diff = a.blockNumber - b.blockNumber;
                  if (diff) {
                    return diff;
                  }
                  return a.transactionIndex - b.transactionIndex;
                });

                //convert to consistent block structure
                byBlock = byBlockAndHash(events);


                log.debug("Retrieved", events.length, "events in", Date.now() - start, "ms");

                _context6.prev = 19;
                blocks = _lodash2.default.values(byBlock);

                log.debug("Sending", blocks.length, "blocks to callback");
                meta = _extends({}, this.meta);
                //for each block

                i = 0;

              case 24:
                if (!(i < blocks.length)) {
                  _context6.next = 37;
                  break;
                }

                b = blocks[i];

                //retrieve receipts if asked to do so. this will be slow...

                if (!this.includeReceipt) {
                  _context6.next = 29;
                  break;
                }

                _context6.next = 29;
                return this._getReceipts(b.transactions);

              case 29:
                _context6.next = 31;
                return cb(null, b.transactions, meta);

              case 31:
                if (!(typeof badCb === 'function')) {
                  _context6.next = 34;
                  break;
                }

                _context6.next = 34;
                return badCb(b.badTransactions);

              case 34:
                ++i;
                _context6.next = 24;
                break;

              case 37:

                this.meta = {
                  rpcCalls: 0
                };
                _context6.next = 43;
                break;

              case 40:
                _context6.prev = 40;
                _context6.t0 = _context6['catch'](19);

                log.error("Problem in callback", _context6.t0);

              case 43:

                //if there is more in the entire block range
                log.debug("Final end", this.finalEnd, "Current end", this.toBlock);
                if (this.finalEnd > this.toBlock) {
                  _start = this.toBlock + 1;

                  if (this.increment < MIN_BLOCK_RANGE) {
                    this.increment = MIN_BLOCK_RANGE;
                  }
                  end = this.toBlock + 1 + this.increment;


                  log.debug("Going to next segement", _start, end);
                  this.fromBlock = _start;
                  this.toBlock = Math.min(end, this.finalEnd);
                  done(this);
                } else {
                  log.debug("Finished all segments");
                  //otherwise scan is complete
                  done(undefined);
                }
                _context6.next = 66;
                break;

              case 47:
                _context6.prev = 47;
                _context6.t1 = _context6['catch'](6);

                if (!TOO_MANY.test(_context6.t1.message)) {
                  _context6.next = 64;
                  break;
                }

                if (!(span <= 1)) {
                  _context6.next = 52;
                  break;
                }

                throw _context6.t1;

              case 52:

                //otherwise, cut the span in 1/2 and try again
                newSpan = Math.ceil(span / 2) - 0;

                //if wec can't split any lower than 1, we bail

                if (!(newSpan === 0)) {
                  _context6.next = 55;
                  break;
                }

                throw _context6.t1;

              case 55:
                totalSpan = this.finalBlock - this.fromBlock;


                this.increment = newSpan;
                this.totalPages = Math.ceil(totalSpan / this.increment);
                oldEnd = this.toBlock;

                this.toBlock = newSpan + this.fromBlock;
                log.warn("Too many events in range", this.fromBlock, "-", oldEnd, "Trying smaller block range of", newSpan, "blocks");
                this._pull(done, err, cb);
                _context6.next = 66;
                break;

              case 64:
                log.error("Problem pulling events", _context6.t1);
                err(_context6.t1);

              case 66:
              case 'end':
                return _context6.stop();
            }
          }
        }, _callee6, this, [[6, 47], [19, 40]]);
      }));

      function _pull(_x9, _x10, _x11, _x12) {
        return _ref6.apply(this, arguments);
      }

      return _pull;
    }() //end _pull

  }, {
    key: '_getReceipts',
    value: function () {
      var _ref7 = _asyncToGenerator( /*#__PURE__*/regeneratorRuntime.mark(function _callee7(txns) {
        var web3, i, r;
        return regeneratorRuntime.wrap(function _callee7$(_context7) {
          while (1) {
            switch (_context7.prev = _context7.next) {
              case 0:
                web3 = this.web3;
                i = 0;

              case 2:
                if (!(i < txns.length)) {
                  _context7.next = 10;
                  break;
                }

                _context7.next = 5;
                return web3.eth.getTransactionReceipt(txns[i].transactionHash);

              case 5:
                r = _context7.sent;

                txns[i].receipt = r;

              case 7:
                ++i;
                _context7.next = 2;
                break;

              case 10:
              case 'end':
                return _context7.stop();
            }
          }
        }, _callee7, this);
      }));

      function _getReceipts(_x13) {
        return _ref7.apply(this, arguments);
      }

      return _getReceipts;
    }()
  }]);

  return EthSyncCursor;
}();

exports.default = EthSyncCursor;

var EthBlock = function () {
  function EthBlock(props) {
    var _this5 = this;

    _classCallCheck(this, EthBlock);

    this.number = props ? props.number : undefined;
    this.timestamp = props ? props.timestamp : undefined;

    this._byHash = {};
    this._badTxns = [];
    ['addEvent'].forEach(function (fn) {
      return _this5[fn] = _this5[fn].bind(_this5);
    });
  }

  _createClass(EthBlock, [{
    key: 'addEvent',
    value: function addEvent(evt) {
      var hash = evt.transactionHash;
      if (!hash) {
        log.debug("Invalid event", JSON.stringify(evt, null, 2));
        this._badTxns.push(evt);
        return;
        //throw new Error("Missing transactionHash in event");
      }

      hash = hash.toLowerCase();
      var bundle = this._byHash[hash];
      if (!bundle) {
        bundle = new EventBundle({
          transactionHash: hash,
          transactionIndex: evt.transactionIndex,
          blockNumber: this.number,
          timestamp: this.timestamp
        });
        this._byHash[hash] = bundle;
      }
      bundle.addEvent(evt);
    }
  }, {
    key: 'transactions',
    get: function get() {
      return _lodash2.default.values(this._byHash);
    }
  }, {
    key: 'badTransactions',
    get: function get() {
      return this._badTxns;
    }
  }, {
    key: 'byHash',
    get: function get() {
      return _extends({}, this._byHash);
    }
  }]);

  return EthBlock;
}();

var EventBundle = function () {
  function EventBundle(props) {
    var _this6 = this;

    _classCallCheck(this, EventBundle);

    this.transactionHash = props.transactionHash;
    this.transactionIndex = props.transactionIndex;
    this.blockNumber = props.blockNumber;
    this.timestamp = props.timestamp;
    this.logEvents = {};
    ['addEvent'].forEach(function (fn) {
      return _this6[fn] = _this6[fn].bind(_this6);
    });
  }

  _createClass(EventBundle, [{
    key: 'addEvent',
    value: function addEvent(evt) {
      var a = this.logEvents[evt.event] || [];
      a.push(evt);
      a.sort(function (a, b) {
        return a.logIndex - b.logIndex;
      });
      this.logEvents[evt.event] = a;
    }
  }]);

  return EventBundle;
}();

var byBlockAndHash = function byBlockAndHash(events) {
  return events.reduce(function (o, e) {
    var retVals = e.returnValues;
    _lodash2.default.keys(retVals).forEach(function (k) {
      var d = retVals[k];
      if (d._ethersType === 'BigNumber') {
        retVals[k] = d.toString();
      }
    });
    var block = o[e.blockNumber] || new EthBlock({
      number: e.blockNumber
    });
    block.addEvent(e);
    o[e.blockNumber] = block;
    return o;
  }, {});
};
//# sourceMappingURL=EthSyncCursor.js.map